@@ -1,16 +1,17 @@
+use crate::directory::Directory;
 use crate::error::{Error, Result};
 use crate::manifest::Name;
 use crate::run::Project;
 use crate::rustflags;
-use serde::Deserialize;
+use serde_derive::Deserialize;
 use std::path::PathBuf;
 use std::process::{Command, Output, Stdio};
-use std::{env, fs};
+use std::{env, fs, iter};
 
 #[derive(Deserialize)]
 pub struct Metadata {
-    pub target_directory: PathBuf,
-    pub workspace_root: PathBuf,
+    pub target_directory: Directory,
+    pub workspace_root: Directory,
     pub packages: Vec<Package>,
 }
 
@@ -29,16 +30,20 @@ fn raw_cargo() -> Command {
 fn cargo(project: &Project) -> Command {
     let mut cmd = raw_cargo();
     cmd.current_dir(&project.dir);
-    cmd.env(
-        "CARGO_TARGET_DIR",
-        path!(project.target_dir / "tests" / "target"),
-    );
+    cmd.envs(cargo_target_dir(project));
+    cmd.envs(rustflags::envs());
     cmd.arg("--offline");
-    rustflags::set_env(&mut cmd);
     cmd
 }
 
-pub fn build_dependencies(project: &Project) -> Result<()> {
+fn cargo_target_dir(project: &Project) -> impl Iterator<Item = (&'static str, PathBuf)> {
+    iter::once((
+        "CARGO_TARGET_DIR",
+        path!(project.target_dir / "tests" / "target"),
+    ))
+}
+
+pub fn build_dependencies(project: &mut Project) -> Result<()> {
     let workspace_cargo_lock = path!(project.workspace / "Cargo.lock");
     if workspace_cargo_lock.exists() {
         let _ = fs::copy(workspace_cargo_lock, path!(project.dir / "Cargo.lock"));
@@ -46,20 +51,30 @@ pub fn build_dependencies(project: &Project) -> Result<()> {
         let _ = cargo(project).arg("generate-lockfile").status();
     }
 
-    let status = cargo(project)
+    let mut command = cargo(project);
+    command
         .arg(if project.has_pass { "build" } else { "check" })
         .args(target())
         .arg("--bin")
         .arg(&project.name)
-        .args(features(project))
-        .status()
-        .map_err(Error::Cargo)?;
+        .args(features(project));
 
-    if status.success() {
-        Ok(())
-    } else {
-        Err(Error::CargoFail)
+    let status = command.status().map_err(Error::Cargo)?;
+    if !status.success() {
+        return Err(Error::CargoFail);
     }
+
+    // Check if this Cargo contains https://github.com/rust-lang/cargo/pull/10383
+    project.keep_going = command
+        .arg("-Zunstable-options")
+        .arg("--keep-going")
+        .stdout(Stdio::null())
+        .stderr(Stdio::null())
+        .status()
+        .map(|status| status.success())
+        .unwrap_or(false);
+
+    Ok(())
 }
 
 pub fn build_test(project: &Project, name: &Name) -> Result<Output> {
@@ -80,6 +95,7 @@ pub fn build_test(project: &Project, name: &Name) -> Result<Output> {
         .args(features(project))
         .arg("--quiet")
         .arg("--color=never")
+        .arg("--message-format=json")
         .output()
         .map_err(Error::Cargo)
 }
