@@ -8,12 +8,15 @@ use crate::manifest::{Bin, Build, Config, Manifest, Name, Package, Workspace};
 use crate::message::{self, Fail, Warn};
 use crate::normalize::{self, Context, Variations};
 use crate::{features, rustflags, Expected, Runner, Test};
+use serde_derive::Deserialize;
 use std::collections::BTreeMap as Map;
 use std::env;
 use std::ffi::{OsStr, OsString};
 use std::fs::{self, File};
 use std::mem;
 use std::path::{Path, PathBuf};
+use std::process::Output;
+use std::str;
 
 #[derive(Debug)]
 pub struct Project {
@@ -28,6 +31,7 @@ pub struct Project {
     pub workspace: Directory,
     pub path_dependencies: Vec<PathDependency>,
     manifest: Manifest,
+    pub keep_going: bool,
 }
 
 #[derive(Debug)]
@@ -42,9 +46,9 @@ impl Runner {
         filter(&mut tests);
 
         let (project, _lock) = (|| {
-            let project = self.prepare(&tests)?;
+            let mut project = self.prepare(&tests)?;
             let lock = Lock::acquire(path!(project.dir / ".lock"));
-            self.write(&project)?;
+            self.write(&mut project)?;
             Ok((project, lock))
         })()
         .unwrap_or_else(|err| {
@@ -146,10 +150,11 @@ impl Runner {
             workspace,
             path_dependencies,
             manifest,
+            keep_going: false,
         })
     }
 
-    fn write(&self, project: &Project) -> Result<()> {
+    fn write(&self, project: &mut Project) -> Result<()> {
         let manifest_toml = toml::to_string(&project.manifest)?;
 
         let config = self.make_config();
@@ -158,7 +163,12 @@ impl Runner {
         fs::create_dir_all(path!(project.dir / ".cargo"))?;
         fs::write(path!(project.dir / ".cargo" / "config"), config_toml)?;
         fs::write(path!(project.dir / "Cargo.toml"), manifest_toml)?;
-        fs::write(path!(project.dir / "main.rs"), b"fn main() {}\n")?;
+
+        let main_rs = b"\
+            #![allow(unknown_lints, unused_crate_dependencies, missing_docs)]\n\
+            fn main() {}\n\
+        ";
+        fs::write(path!(project.dir / "main.rs"), &main_rs[..])?;
 
         cargo::build_dependencies(project)?;
 
@@ -259,7 +269,9 @@ impl Test {
         message::begin_test(self, show_expected);
         check_exists(&self.path)?;
 
-        let output = cargo::build_test(project, name)?;
+        let mut output = cargo::build_test(project, name)?;
+        let src_path = project.source_dir.join(&self.path);
+        parse_cargo_json(&src_path, &mut output);
         let success = output.status.success();
         let stdout = output.stdout;
         let stderr = normalize::diagnostics(
@@ -481,3 +493,61 @@ fn filter(tests: &mut Vec<ExpandedTest>) {
             .any(|f| t.test.path.to_string_lossy().contains(f))
     });
 }
+
+#[derive(Deserialize)]
+struct CargoMessage {
+    #[allow(dead_code)]
+    reason: Reason,
+    target: RustcTarget,
+    message: RustcMessage,
+}
+
+#[derive(Deserialize)]
+enum Reason {
+    #[serde(rename = "compiler-message")]
+    CompilerMessage,
+}
+
+#[derive(Deserialize)]
+struct RustcTarget {
+    src_path: PathBuf,
+}
+
+#[derive(Deserialize)]
+struct RustcMessage {
+    rendered: String,
+    level: String,
+}
+
+fn parse_cargo_json(src_path: &Path, output: &mut Output) {
+    let mut diagnostics = String::new();
+    let mut nonmessage_stdout = String::new();
+    let mut remaining = match str::from_utf8(&output.stdout) {
+        Ok(remaining) => remaining,
+        // TODO: this should be easy to support if someone needs it, just needs
+        // an alternative to `find` below.
+        Err(_) => return,
+    };
+    while !remaining.is_empty() {
+        let begin = match remaining.find("{\"reason\":") {
+            Some(begin) => begin,
+            None => break,
+        };
+        let (nonmessage, rest) = remaining.split_at(begin);
+        nonmessage_stdout.push_str(nonmessage);
+        let len = match rest.find('\n') {
+            Some(end) => end + 1,
+            None => rest.len(),
+        };
+        let (message, rest) = rest.split_at(len);
+        if let Ok(de) = serde_json::from_str::<CargoMessage>(message) {
+            if de.message.level != "failure-note" && de.target.src_path == src_path {
+                diagnostics.push_str(&de.message.rendered);
+            }
+        }
+        remaining = rest;
+    }
+    nonmessage_stdout.push_str(remaining);
+    output.stdout = Vec::from(nonmessage_stdout);
+    output.stderr = Vec::from(diagnostics);
+}
